/*
  maccasserver.c
  B-CAS server program for BonCasLink protocol

  Copyright 2010 tmkk. All rights reserved.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pthread.h>
#include <arpa/inet.h>

#include <PCSC/winscard.h>
#include <PCSC/wintypes.h>

#define VERSION 20100504

#define CHECK(f, rv) \
if (SCARD_S_SUCCESS != rv) \
{ \
fprintf(stderr, f ": %s\n", pcsc_stringify_error(rv)); \
return 0; \
}

static const unsigned char initCmd[5] = {0x90, 0x30, 0x00, 0x00, 0x00};
SCARDHANDLE hCard;

void* communicationThread(void* pParam)
{
	unsigned char recvBuf[256], sendBuf[256];
	int recvSize, fd=(int)pParam, rv;
	unsigned int bufSize;
	
	while(1)
	{
		recvSize = recv(fd,recvBuf,256,0);
		
		if(recvSize == 0 || !recvBuf[0]) {
			close(fd);
			break;
		}
		bufSize = 255;
		rv = SCardTransmit(hCard, SCARD_PCI_T1, recvBuf+1, recvBuf[0], NULL, sendBuf+1, &bufSize);
		if(SCARD_S_SUCCESS != rv) {
			fprintf(stderr,"SCardTransmit failed, closing socket.\n");
			close(fd);
			break;
		}
		sendBuf[0] = bufSize;
		send(fd,sendBuf,bufSize+1,0);
	}
	
	fprintf(stderr,"Connection closed.\n");
	return NULL;
}


int main(int argc, char *argv[])
{
	int rv, yes=1, i;
	SCARDCONTEXT hContext;
	char *readerName = NULL;
	unsigned int ret, bufSize;
	unsigned char buf[256];
	int sock_serv, port=6900;
	struct sockaddr_in addr;
	
	for(i=1;i<argc;i++) {
		if(!strncmp(argv[i],"-p",2) && i < argc-1) {
			port = atoi(argv[++i]);
		}
		else if(!strncmp(argv[i],"-r",2) && i < argc-1) {
			readerName = (char *)malloc(strlen(argv[i+1]));
			strcpy(readerName,argv[++i]);
		}
		else if(!strncmp(argv[i],"-h",2)) {
			fprintf(stderr,"MacCasServer %d by tmkk\nUsage: %s [-p ListenPort] [-r ReaderName]\n",VERSION,argv[0]);
			return 0;
		}
	}
	
	rv = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	CHECK("SCardEstablishContext", rv);
	
	if(!readerName) {
		rv = SCardListReaders(hContext, NULL, NULL, &ret);
		CHECK("SCardListReaders", rv);
		
		readerName = (char *)calloc(ret, sizeof(char));
		rv = SCardListReaders(hContext, NULL, readerName, &ret);
		CHECK("SCardListReaders", rv);
		
		if(ret > strlen(readerName)+2) {
			fprintf(stderr,"Found multiple card readers. Please specify the preferred one with -r option.\n");
			fprintf(stderr,"Available readers:\n");
			char *ptr = readerName;
			while(1) {
				fprintf(stderr,"\t\"%s\"\n",ptr);
				ret -= strlen(ptr)+1;
				ptr += strlen(ptr)+1;
				if(!ret || !*ptr) break;
			}
			free(readerName);
			SCardReleaseContext(hContext);
			return 0;
		}
		
		fprintf(stderr, "Detected card reader: %s\n", readerName);
	}
	
	rv = SCardConnect(hContext, readerName, SCARD_SHARE_SHARED, SCARD_PROTOCOL_T1, &hCard, &ret);
	CHECK("SCardConnect", rv);
	
	if(ret != SCARD_PROTOCOL_T1) {
		fprintf(stderr,"Invalid available protocol.\n");
		goto last;
	}
	
	bufSize = 256;
	rv = SCardStatus(hCard, NULL, &ret, NULL, NULL, buf, &bufSize);
	CHECK("SCardStatus", rv);
	fprintf(stderr,"Card opened, ATR: ");
	for(i=0;i<bufSize;i++) fprintf(stderr,"%02x ",buf[i]);
	fprintf(stderr,"\n");
	
	bufSize = 256;
	rv = SCardTransmit(hCard, SCARD_PCI_T1, initCmd, sizeof(initCmd), NULL, buf, &bufSize);
	CHECK("SCardTransmit", rv);
	
	if(bufSize < 57) {
		fprintf(stderr,"Card init failed. Maybe \n");
		goto last;
	}
	fprintf(stderr,"Found B-CAS Card, ID: %02x-%02x-%02x-%02x-%02x-%02x\n",buf[8],buf[9],buf[10],buf[11],buf[12],buf[13]);
	
	sock_serv = socket(AF_INET, SOCK_STREAM, 0);
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = INADDR_ANY;
	
	setsockopt(sock_serv, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(yes));
	
	if(bind(sock_serv, (struct sockaddr *)&addr, sizeof(addr)) != 0) {
		perror("bind");
		goto last;
	}
	if(listen(sock_serv, 5) != 0) {
		perror("listen");
		goto last;
	}
	
	fprintf(stderr,"Listening on port %d...\n", ntohs(addr.sin_port));
	
	while(1) {
		pthread_t tid;
		struct sockaddr_in client;
		unsigned int len = sizeof(client);
		int sock_cli = accept(sock_serv, (struct sockaddr *)&client, &len);
		if(sock_cli < 0) {
			perror("accept");
			break;
		}
		fprintf(stderr,"Accepted connection from %s:%d\n",inet_ntoa(client.sin_addr), ntohs(client.sin_port));
		if(pthread_create(&tid, NULL, (void *)communicationThread, (void *)sock_cli)!= 0) {
			perror("pthread_create");
			close(sock_cli);
			break;
		}
		pthread_detach(tid);
	}
	
	close(sock_serv);
	
  last:
	rv = SCardDisconnect(hCard, SCARD_LEAVE_CARD);
	CHECK("SCardDisconnect", rv);
	
	free(readerName);
	
	rv = SCardReleaseContext(hContext);
	CHECK("SCardReleaseContext", rv);
	
	return 0;
}